#=====================================================================
# SQL-Ledger Accounting
# Copyright (C) 2001
#
#  Author: Dieter Simader
#   Email: dsimader@sql-ledger.org
#     Web: http://www.sql-ledger.org
#
#  Contributors: Thomas Bayen <bayen@gmx.de>
#                Antti Kaihola <akaihola@siba.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#======================================================================
# Utilities for parsing forms
# and supporting routines for linking account numbers
# used in AR, AP and IS modules
#
# CHANGE LOG:
#   DS. 2000-07-04  Created
#   TB. 2000-07-17  Added package Locale
#   AK. 2001-05-14  Added query string for error reporting
#   DS. 2001-06-12  Changed relations from accno to chart_id
#   BL. 2001-06-17  Added style sheet
#
#======================================================================

package Form;


sub new {

  my $type = shift;
  my $self = {};

  read(STDIN, $_, $ENV{CONTENT_LENGTH});
  
  if ($ENV{QUERY_STRING}) {
    $_ = $ENV{QUERY_STRING};
  }

  if ($ARGV[0]) {
    $_ = $ARGV[0];
  }
  
  foreach $item (split(/&/)) {
    ($key, $value) = split(/=/, $item);
    $self->{$key} = &unescape("",$value);
  }

  $self->{action} = lc $self->{action};
  $self->{action} =~ s/( |-)/_/g;
  
  bless $self, $type;
  
}


sub escape {
  my ($self, $str) = @_;

  $str =~ s/([^a-zA-Z0-9_.-])/uc sprintf("%%%02x", ord($1))/ge;
  
  $str;

}


sub unescape {
  my ($self, $str) = @_;
  
  $str =~ tr/+/ /;
  $str =~ s/%([0-9a-fA-Z]{2})/pack("c",hex($1))/eg;
  
  $str;

}


sub error {
  my ($self, $msg) = @_;

  if ($ENV{HTTP_USER_AGENT}) {
    print qq|Content-Type: text/html

    <body bgcolor=ffffff>

    <h2><font color=red>Error!</font></h2>

    <p><b>$msg</b>|;

    die "Error: $msg\n";

  } else {
  
    if ($self->{error_function}) {
      &{ $self->{error_function} }($msg);
    } else {
      die "Error: $msg\n";
    }
  }
  
}


sub dberror {
  my ($self, $msg) = @_;

  $self->error(DBI::errstr . " $msg");
  
}


sub isblank {
  my ($self, $name, $msg) = @_;

  if ($self->{$name} =~ /^\s*$/) {
    $self->error($msg);
  }
}
  

sub header {
  my $self = shift;

  my ($nocache, $stylesheet, $charset);
  
  # Opera caches scripts
  # use expire tag to prevent caching
  if ($ENV{HTTP_USER_AGENT} =~ /Opera/) {
    $nocache = qq|<META HTTP-EQUIV="Expires" CONTENT="Tue, 01 Jan 1980 1:00:00 GMT">
  <META HTTP-EQUIV="Pragma" CONTENT="no-cache">
|;
  }

  if ($self->{stylesheet} && (-f $self->{stylesheet})) {
    $stylesheet = qq|<LINK REL="stylesheet" HREF="$self->{stylesheet}" TYPE="text/css" TITLE="SQL-Ledger style sheet">
|;
  }

  if ($self->{charset}) {
    $charset = qq|<META HTTP-EQUIV="Charset" CONTENT="$self->{charset}">
|;
  }

  print qq|Content-Type: text/html

<head>
  <title>$self->{title}</title>
  $nocache
  $stylesheet
  $charset
</head>

<body bgcolor=ffffff>
|;

}


sub redirect {
  my $self = shift;

  if ($self->{callback}) {

    ($script, $argv) = split(/\?/, $self->{callback});

    exec "./$script", $argv;
    
  } else {
    if ($ENV{HTTP_USER_AGENT}) {
      $self->header;
    }
  }
}


sub isposted {
  my ($self, $rc) = @_;

  if ($rc) {
    $self->redirect;
  }

  $rc;
  
}


sub isdeleted {
  my ($self, $rc) = @_;

  if ($rc) {
    $self->redirect;
  }

  $rc;
  
}


sub sort_columns {
  my ($self, @columns) = @_;

  @columns = grep !/^$self->{sort}$/, @columns;
  splice @columns, 0, 0, $self->{sort};

  @columns;
  
}


sub format_amount {
  my ($self, $myconfig, $amount, $dash, $vg) = @_;

  $amount = ($vg) ? $amount : sprintf("%.2f", $amount);
  my $negative = ($amount < 0);
  
  if ($amount != 0) {
    if ($myconfig->{numberformat} && ($myconfig->{numberformat} ne '1000.00')) {
      my ($whole, $dec) = split /\./, "$amount";
      $whole =~ s/-//;
      $amount = join '', reverse split //, $whole;
      
      if ($myconfig->{numberformat} eq '1,000.00') {
	$amount =~ s/\d{3,}?/$&,/g;
	$amount =~ s/,$//;
	$amount = join '', reverse split //, $amount;
	$amount .= "\.$dec" if $dec;
      }
      
      if ($myconfig->{numberformat} eq '1.000,00') {
	$amount =~ s/\d{3,}?/$&./g;
	$amount =~ s/\.$//;
	$amount = join '', reverse split //, $amount;
	$amount .= ",$dec" if $dec;
      }
      
      if ($myconfig->{numberformat} eq '1000,00') {
	$amount = "$whole";
	$amount .= ",$dec" if $dec;
      }

      if ($negative) {
	$amount = ($dash) ? "($amount)" : "-$amount";
      }
    }
  } else {
    $amount = ($dash) ? "- " : "";
  }

  $amount;

}


sub parse_amount {
  my ($self, $myconfig, $amount) = @_;

  if (($myconfig->{numberformat} eq '1.000,00') ||
      ($myconfig->{numberformat} eq '1000,00')) {
    $amount =~ s/\.//g;
    $amount =~ s/,/\./;
  }

  $amount =~ s/,//g;
  
  return ($amount * 1);

}


sub parse_template {
  my $self = shift;

  open(IN, "$self->{templates}/$self->{IN}") || $self->error("$!: $self->{IN}");

  @_ = <IN>;
  close(IN);
  
  if ($self->{OUT}) {
    open(OUT, "$self->{OUT}") || $self->error("$!: $self->{OUT}");
  } else {
    open(OUT, ">-") || $self->error("$!: STDOUT");
  }
  
  # read file and replace <%variable%>
  while ($_ = shift) {
      
    $par = "";
    $var = $_;

    if (/<%foreach /) {
      
      # this one we need for the count
      chomp $var;
      $var =~ s/<%foreach (.+?)%>/$1/;
      while ($_ = shift) {
	last if (/<%end /);

	# store line in $par
	$par .= $_;
      }
      
      # display contents of $self->{number}[] array
      for $i (0 .. $#{ $self->{$var} }) {
	# don't parse par, we need it for each line
	$_ = $par;
	s/<%(.+?)%>/$self->{$1}[$i]/mg;
	print OUT;
      }
      next;
    }
    
    if (/<%if not /) {
      # check if it is not set and display
      chop;
      s/<%if not (.+?)%>/$1/;

      unless ($self->{$_}) {
	while ($_ = shift) {
	  last if (/<%end /);

	  # store line in $par
	  $par .= $_;
	}
	
	$_ = $par;
	
      } else {
	while ($_ = shift) {
	  last if (/<%end /);
	}
	next;
      }
    }


    if (/<%if /) {
      # check if it is set and display
      chop;
      s/<%if (.+?)%>/$1/;

      if ($self->{$_}) {
	while ($_ = shift) {
	  last if (/<%end /);

	  # store line in $par
	  $par .= $_;
	}
	
	$_ = $par;
	
      } else {
	while ($_ = shift) {
	  last if (/<%end /);
	}
	next;
      }
    }
    
    # check for <%include filename%>
    if (/<%include /) {
      
      # get the filename
      chomp $var;
      $var =~ s/<%include (.+?)%>/$1/;

      # mangle filename
      $var =~ s/\///g;

      # prevent the infinite loop!
      next if ($self->{"$var"});

      open(INC, "$self->{templates}/$var") || $self->error("$!: $self->{templates}/$var");
      unshift(@_, <INC>);
      close(INC);

      $self->{"$var"} = 1;

      next;
    }
    
    s/<%(.+?)%>/$self->{$1}/g;
    print OUT;
  }

  close(OUT);

}



# Database routines used throughout

sub dbconnect {
  my ($self, $myconfig) = @_;

  # connect to database
  my $dbh = DBI->connect($myconfig->{dbconnect}, $myconfig->{dbuser}, $myconfig->{dbpasswd}) || $self->dberror;

  # set db options
  if ($myconfig->{dboptions}) {
    $dbh->do($myconfig->{dboptions}) || $self->dberror($myconfig->{dboptions});
  }

  $dbh;

}


sub dbconnect_noauto {
  my ($self, $myconfig) = @_;

  # connect to database
  $dbh = DBI->connect($myconfig->{dbconnect}, $myconfig->{dbuser}, $myconfig->{dbpasswd}, {AutoCommit => 0}) || $self->dberror;

  # set db options
  if ($myconfig->{dboptions}) {
    $dbh->do($myconfig->{dboptions}) || $self->dberror($myconfig->{dboptions});
  }

  $dbh;

}


sub update_balance {
  my ($self, $dbh, $table, $field, $where, $value) = @_;

  # if we have a value, go do it
  if ($value != 0) {
    # retrieve balance from table
    my $query = "SELECT $field FROM $table WHERE $where";
    my $sth = $dbh->prepare($query);

    $sth->execute || $self->dberror($query);
    my ($balance) = $sth->fetchrow_array;
    $sth->finish;

    $balance += $value;
    # update balance
    $query = "UPDATE $table SET $field = $balance WHERE $where";
    $dbh->do($query) || $self->dberror($query);
  }
}



# the selection sub is used in the AR, AP and IS module
#
sub all_vc {
  my ($self, $myconfig, $table) = @_;

  # create array for vendor or customer
  my $dbh = $self->dbconnect($myconfig);

  my $query = qq|SELECT id, name
                 FROM $table
	         ORDER BY name|;
  my $sth = $dbh->prepare($query);
  $sth->execute || $self->dberror($query);
  
  my $ref = $sth->fetchrow_hashref;
  push @{ $self->{"all_$table"} }, $ref;

  $self->{"${table}_id"} = $ref->{id};

  while ($ref = $sth->fetchrow_hashref) {
    push @{ $self->{"all_$table"} }, $ref;
  }

  $sth->finish;
  $dbh->disconnect;

}


sub create_links {
  my ($self, $module, $myconfig, $table) = @_;

  # get all the customers or vendors
  &all_vc($self, $myconfig, $table);

  my $dbh = $self->dbconnect($myconfig);
  # now get the account numbers
  my $query = qq|SELECT accno, description, link
                 FROM chart
		 WHERE link LIKE '%$module%'
		 ORDER BY accno|;
  my $sth = $dbh->prepare($query);
  $sth->execute || $self->dberror($query);

  while (my $ref = $sth->fetchrow_hashref) {
    foreach my $key (split(/:/, $ref->{link})) {
      if ($key =~ /$module/) {
	push @{ $self->{"${module}_links"}{$key} }, { accno => $ref->{accno},
                                       description => $ref->{description} };
      }
    }
  }
  $sth->finish;

  
  # get account numbers and value
  if ($self->{id}) {
    if ($table eq "vendor") {
      $query = qq|SELECT invnumber, transdate, vendor as vendor_id, datepaid,
		  duedate, ordnumber, taxincluded
		  FROM ap
		  WHERE id = $self->{id}|;
    }
    if ($table eq "customer") {
      $query = qq|SELECT invnumber, transdate, customer as customer_id,
                  datepaid, duedate, taxincluded
		  FROM ar
		  WHERE id = $self->{id}|;
    }

    $sth = $dbh->prepare($query);
    $sth->execute || $self->dberror($query);
    
    $ref = $sth->fetchrow_hashref;
    foreach $key (keys %$ref) {
      $self->{$key} = $ref->{$key};
    }
    $sth->finish;
    
    # get amounts from individual entries
    $query = qq|SELECT chart.accno, amount
		FROM acc_trans, chart
		WHERE chart.id = acc_trans.chart_id
		AND trans_id = $self->{id}|;
    $sth = $dbh->prepare($query);
    $sth->execute || $self->dberror($query);

    # store amounts in {amount}{$accno}
    while ($ref = $sth->fetchrow_hashref) {
      $self->{amount}{$ref->{accno}} = $ref->{amount};
    }
    $sth->finish;

  } else {
    # get date
    $query = qq|SELECT current_date, current_date + 30|;
    $sth = $dbh->prepare($query);
    $sth->execute || $self->dberror($query);

    ($self->{transdate}, $self->{duedate}) = $sth->fetchrow_array;
    $sth->finish;
  }

  $dbh->disconnect;

}


sub current_date {
  my ($self, $myconfig, $thisdate, $days) = @_;
  
  my $dbh = $self->dbconnect($myconfig);
  my $query = qq|SELECT current_date|;

  $days *= 1;
  if ($thisdate) {
    $query = qq|SELECT date '$thisdate' + $days|;
  }
  
  my $sth = $dbh->prepare($query);
  $sth->execute || $self->dberror($query);

  ($thisdate) = $sth->fetchrow_array;
  $sth->finish;

  $dbh->disconnect;

  $thisdate;

}



package Locale;


sub new {
  my ($type, $country, $NLS_file) = @_;
  my $self = {};

  %self = ();
  if ($country) {
    if (-d "locale/$country") {
      $self->{countrycode} = $country;
      
      if (-e "locale/$country/$NLS_file") {
	require "locale/$country/$NLS_file";
      }
    }
  }

  $self->{NLS_file} = $NLS_file;
  
  push @{ $self->{LONG_MONTH} }, (qw(January February March April May June July August September October November December));
  push @{ $self->{SHORT_MONTH} }, (qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec));
  
  bless $self, $type;

}


sub text {
  my ($self, $text) = @_;
  
  return (exists $self{texts}{$text}) ? $self{texts}{$text} : $text;
  
}


sub findsub {
  my ($self, $text) = @_;

  if (exists $self{subs}{$text}) {
    $text = $self{subs}{$text};
  } else {
    if ($self->{countrycode} && $self->{NLS_file}) {
      Form->error("Edit locale/$self->{countrycode}/$self->{NLS_file} and add the variable ($text) in the \$self{subs} section!");
    }
  }

  $text;

}


sub date {
  my ($self, $myconfig, $date, $longformat) = @_;

  my $longdate = "";
  my $longmonth = ($longformat) ? 'LONG_MONTH' : 'SHORT_MONTH';
  
  if ($date) {
    $spc = substr($myconfig->{dateformat}, 2, 1);
    if ($spc eq '.') {
      $spc = '\.';
    }
    if ($spec eq '/') {
      $spc = '\/';
    }

    if ($myconfig->{dateformat} eq 'yy-mm-dd') {
      ($yy, $mm, $dd) = split /$spc/, $date;
      $dd *= 1;
      $mm--;
      $yy = ($yy < 70) ? $yy + 2000 : $yy;
      $longdate = &text($self, $self->{$longmonth}[$mm])." $dd, $yy";
    }
    if ($myconfig->{dateformat} =~ /^mm/) {
      ($mm, $dd, $yy) = split /$spc/, $date;
      $dd *= 1;
      $mm--;
      $yy = ($yy < 70) ? $yy + 2000 : $yy;
      $longdate = &text($self, $self->{$longmonth}[$mm])." $dd, $yy";
    }
    if ($myconfig->{dateformat} =~ /^dd/) {
      ($dd, $mm, $yy) = split /$spc/, $date;
      $dd *= 1;
      $mm--;
      $yy = ($yy < 70) ? $yy + 2000 : $yy;
      $longdate = "$dd ".&text($self, $self->{$longmonth}[$mm]).", $yy";
    }
  }

  $longdate;

}


1;

